/* USER CODE BEGIN Header */
/**
  ******************************************************************************
  * @file           : main.c
  * @brief          : Main program body
  ******************************************************************************
  * @attention
  *
  * Copyright (c) 2023 STMicroelectronics.
  * All rights reserved.
  *
  * This software is licensed under terms that can be found in the LICENSE file
  * in the root directory of this software component.
  * If no LICENSE file comes with this software, it is provided AS-IS.
  *
  ******************************************************************************
  */
/* USER CODE END Header */
/* Includes ------------------------------------------------------------------*/
#include "main.h"
#include "usb_host.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */
typedef
		struct	{
	uint8_t			address;	// アドレスNo.
	uint32_t		bit;		// ビット位置
}
KEYB_TABLE;

#include "usbh_hid.h"
#include "usb_keytable.h"
/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */
typedef
	enum
	{
		PC88_KEYBOARD = 0,
		PC88_MOUSE,
		PC88_GAMEPAD
	}
PC88_USB_MODE;

enum{
	SS_L100 = 0,
	SS_RLDU,
	SS_SACB,
	SS_RXYZ,
	SS_MAX
};

enum
{
	BUTTON_UP = 0,
	BUTTON_DOWN,
	BUTTON_LEFT,
	BUTTON_RIGHT,
	BUTTON_A,
	BUTTON_B,
	BUTTON_C,
	BUTTON_S,
	BUTTON_LR,
	BUTTON_MAX
};

enum
{
	GAMEPAD_MODE1 = 0,
	GAMEPAD_MODE2,
	GAMEPAD_MODE3,
	GAMEPAD_MODE4,
	mode_gamepadMAX
};
/* USER CODE END PTD */

/* Private define ------------------------------------------------------------*/
/* USER CODE BEGIN PD */
#define    KEYB_CTRL       (0xe0)
#define    KEYB_SHIFT      (0xe1)
#define    KEYB_GRAPH      (0xe2)
#define    KEYB_RALT       (0xe6)
#define    KEYB_HYPHEN     (0x87)
#define    KEYB_YEN        (0x89)
#define    KEYB_2          (0x5a)
#define    KEYB_4          (0x5c)
#define    KEYB_6          (0x5e)
#define    KEYB_8          (0x60)
#define    KEYB_X          (0x1b)
#define    KEYB_Z          (0x1d)
#define    KEYB_SPC        (0x2c)
#define    KEYB_RET        (0x28)
#define    KEYB_KANA       (0xe6)
#define    KEYB_ESC        (0x29)
#define    KEYB_STOP       (0x35)
#define    KEYB_UP         (0x52)
#define    KEYB_DOWN       (0x51)
#define    KEYB_LEFT       (0x50)
#define    KEYB_RIGHT      (0x4f)

#define    ADDRESS_MAX     (0x10)

#define MOUSE_UP		(0xF8)
#define MOUSE_CENTER	(0x80)
#define MOUSE_DOWN		(0x08)

// FAST 15shots/sec, SLOW 7.5shots/sec
#define    RAPID_FAST      (2u)
#define    RAPID_SLOW      (4u)
#define    RAPID_COUNT_MAX (100000u)

#define		SS_DELAY_TIME		(4u)
#define		SS_WAIT_TIME		(64u)
/* USER CODE END PD */

/* Private macro -------------------------------------------------------------*/
/* USER CODE BEGIN PM */
#define		SS_DELAY		for(volatile uint32_t dt=0;dt<SS_DELAY_TIME;dt++)
#define		SS_WAIT			for(volatile uint32_t dt=0;dt<SS_WAIT_TIME;dt++)
/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/
TIM_HandleTypeDef htim11;

/* USER CODE BEGIN PV */
static uint32_t					rapid_count = 0;
static uint32_t					rapid_speed = RAPID_FAST;

static KEYB_TABLE				usb_keytable[USB_SIZE];
uint32_t						pc88_data[ADDRESS_MAX] = {0};
uint32_t						pc88_reset = 0x00000000;

static PC88_USB_MODE           pc88_usb_mode = PC88_KEYBOARD;

static HID_KEYBD_Info_TypeDef	*keyb_info = NULL;
static HID_MOUSE_Info_TypeDef	*mouse_info = NULL;
static uint32_t					mode_gamepad = GAMEPAD_MODE1;
static uint32_t					led_list[mode_gamepadMAX] = { LED1_Pin, LED2_Pin, LED3_Pin, LED4_Pin };

const static uint16_t			gamepad_map[mode_gamepadMAX][BUTTON_MAX] =
{
	//    [Up]   [Down]     [Left]     [Right]        [A]         [B]       [C]         [START]       [L+R]
	// NORMAL-ZX
	{ KEYB_8,	KEYB_2,		KEYB_4,		KEYB_6,		KEYB_SPC,	KEYB_Z,		KEYB_X,		KEYB_RET, KEYB_ESC     },
	// NORMAL
	{ KEYB_8,	KEYB_2,		KEYB_4,		KEYB_6,		KEYB_SPC,	KEYB_Z,		KEYB_X,		KEYB_RET, KEYB_STOP    },
	// PC-8001 BUG FIRE
	{ KEYB_8,	KEYB_2,		KEYB_4,		KEYB_6,		KEYB_SPC,	KEYB_X,		KEYB_SHIFT,	KEYB_RET, KEYB_STOP    },
	// PC-8001 走れ！スカイライン
	{ KEYB_8,	KEYB_2,		KEYB_4,		KEYB_6,		KEYB_SPC,	KEYB_Z,		KEYB_SHIFT,	KEYB_RET, KEYB_STOP    },
};
/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
void SystemClock_Config(void);
static void MX_GPIO_Init(void);
static void MX_TIM11_Init(void);
void MX_USB_HOST_Process(void);

/* USER CODE BEGIN PFP */
static void set_usb_mouse(void);
static void set_usb_keycode(void);
static void saturn_gamepad(void);
/* USER CODE END PFP */

/* Private user code ---------------------------------------------------------*/
/* USER CODE BEGIN 0 */

/* USER CODE END 0 */

/**
  * @brief  The application entry point.
  * @retval int
  */
int main(void)
{
  /* USER CODE BEGIN 1 */

  /* USER CODE END 1 */

  /* MCU Configuration--------------------------------------------------------*/

  /* Reset of all peripherals, Initializes the Flash interface and the Systick. */
  HAL_Init();

  /* USER CODE BEGIN Init */

  /* USER CODE END Init */

  /* Configure the system clock */
  SystemClock_Config();

  /* USER CODE BEGIN SysInit */

  /* USER CODE END SysInit */

  /* Initialize all configured peripherals */
  MX_GPIO_Init();
  MX_USB_HOST_Init();
  MX_TIM11_Init();
  /* USER CODE BEGIN 2 */
	// VBUS Disabled
	USB_OTG_FS->GCCFG |= USB_OTG_GCCFG_NOVBUSSENS;

  	// JTAG Disabled
  	GPIOB->MODER = (GPIOB->MODER & 0xFFFFFC3F) | 0x00000140;

	// SATURN TH/TR Start
	HAL_GPIO_WritePin(GPIOB, SS_TH_Pin, GPIO_PIN_SET);
	HAL_GPIO_WritePin(GPIOB, SS_TR_Pin, GPIO_PIN_SET);

	// デフォルトデータ
	pc88_reset = (uint32_t)(PC88_D0_Pin|PC88_D1_Pin|PC88_D2_Pin|PC88_D3_Pin|PC88_D4_Pin|PC88_D5_Pin|PC88_D6_Pin|PC88_D7_Pin);

	// ビットシフト時間短縮のため事前に並べ替えておく
	for( uint32_t index = 0; index < USB_SIZE; index++ ){
		usb_keytable[index].address = pc88_usb_keytable[index].address;
		usb_keytable[index].bit = 0x00000000;
		usb_keytable[index].bit |= (pc88_usb_keytable[index].bit & 0x01 ? ((uint32_t)PC88_D0_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (pc88_usb_keytable[index].bit & 0x02 ? ((uint32_t)PC88_D1_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (pc88_usb_keytable[index].bit & 0x04 ? ((uint32_t)PC88_D2_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (pc88_usb_keytable[index].bit & 0x08 ? ((uint32_t)PC88_D3_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (pc88_usb_keytable[index].bit & 0x10 ? ((uint32_t)PC88_D4_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (pc88_usb_keytable[index].bit & 0x20 ? ((uint32_t)PC88_D5_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (pc88_usb_keytable[index].bit & 0x40 ? ((uint32_t)PC88_D6_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (pc88_usb_keytable[index].bit & 0x80 ? ((uint32_t)PC88_D7_Pin << 16u) : 0x00000000);
	}

	HAL_Delay(200);

	// Start Timer
	HAL_TIM_Base_Start_IT(&htim11);
  /* USER CODE END 2 */

  /* Infinite loop */
  /* USER CODE BEGIN WHILE */
  while (1)
  {
    /* USER CODE END WHILE */
    MX_USB_HOST_Process();

    /* USER CODE BEGIN 3 */
  }
  /* USER CODE END 3 */
}

/**
  * @brief System Clock Configuration
  * @retval None
  */
void SystemClock_Config(void)
{
  RCC_OscInitTypeDef RCC_OscInitStruct = {0};
  RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};

  /** Configure the main internal regulator output voltage
  */
  __HAL_RCC_PWR_CLK_ENABLE();
  __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE2);

  /** Initializes the RCC Oscillators according to the specified parameters
  * in the RCC_OscInitTypeDef structure.
  */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSE;
  RCC_OscInitStruct.HSEState = RCC_HSE_ON;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
  RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSE;
  RCC_OscInitStruct.PLL.PLLM = 25;
  RCC_OscInitStruct.PLL.PLLN = 336;
  RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV4;
  RCC_OscInitStruct.PLL.PLLQ = 7;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    Error_Handler();
  }

  /** Initializes the CPU, AHB and APB buses clocks
  */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV2;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_2) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
  * @brief TIM11 Initialization Function
  * @param None
  * @retval None
  */
static void MX_TIM11_Init(void)
{

  /* USER CODE BEGIN TIM11_Init 0 */

  /* USER CODE END TIM11_Init 0 */

  /* USER CODE BEGIN TIM11_Init 1 */

  /* USER CODE END TIM11_Init 1 */
  htim11.Instance = TIM11;
  htim11.Init.Prescaler = 83;
  htim11.Init.CounterMode = TIM_COUNTERMODE_UP;
  htim11.Init.Period = 16665;
  htim11.Init.ClockDivision = TIM_CLOCKDIVISION_DIV1;
  htim11.Init.AutoReloadPreload = TIM_AUTORELOAD_PRELOAD_ENABLE;
  if (HAL_TIM_Base_Init(&htim11) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN TIM11_Init 2 */

  /* USER CODE END TIM11_Init 2 */

}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};

  /* GPIO Ports Clock Enable */
  __HAL_RCC_GPIOH_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();
  __HAL_RCC_GPIOB_CLK_ENABLE();

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOA, PC88_D0_Pin|PC88_D1_Pin|PC88_D2_Pin|PC88_D3_Pin
                          |PC88_D4_Pin|PC88_D5_Pin|PC88_D6_Pin|PC88_D7_Pin, GPIO_PIN_SET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(LED1_GPIO_Port, LED1_Pin, GPIO_PIN_SET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOB, LED2_Pin|LED3_Pin|LED4_Pin|SS_TH_Pin
                          |SS_TR_Pin, GPIO_PIN_RESET);

  /*Configure GPIO pins : PC88_D0_Pin PC88_D1_Pin PC88_D2_Pin PC88_D3_Pin
                           PC88_D4_Pin PC88_D5_Pin PC88_D6_Pin PC88_D7_Pin */
  GPIO_InitStruct.Pin = PC88_D0_Pin|PC88_D1_Pin|PC88_D2_Pin|PC88_D3_Pin
                          |PC88_D4_Pin|PC88_D5_Pin|PC88_D6_Pin|PC88_D7_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_OD;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

  /*Configure GPIO pins : LED1_Pin LED2_Pin LED3_Pin LED4_Pin */
  GPIO_InitStruct.Pin = LED1_Pin|LED2_Pin|LED3_Pin|LED4_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /*Configure GPIO pins : SS_U_Pin SS_D_Pin SS_L_Pin SS_R_Pin */
  GPIO_InitStruct.Pin = SS_U_Pin|SS_D_Pin|SS_L_Pin|SS_R_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /*Configure GPIO pin : SWITCH_Pin */
  GPIO_InitStruct.Pin = SWITCH_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(SWITCH_GPIO_Port, &GPIO_InitStruct);

  /*Configure GPIO pins : SS_TH_Pin SS_TR_Pin */
  GPIO_InitStruct.Pin = SS_TH_Pin|SS_TR_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /*Configure GPIO pin : PC88_A_Pin */
  GPIO_InitStruct.Pin = PC88_A_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_IT_RISING;
  GPIO_InitStruct.Pull = GPIO_PULLDOWN;
  HAL_GPIO_Init(PC88_A_GPIO_Port, &GPIO_InitStruct);

  /*Configure GPIO pins : PC88_A0_Pin PC88_A1_Pin PC88_A2_Pin PC88_A3_Pin */
  GPIO_InitStruct.Pin = PC88_A0_Pin|PC88_A1_Pin|PC88_A2_Pin|PC88_A3_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_IT_FALLING;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /* EXTI interrupt init*/
  HAL_NVIC_SetPriority(EXTI9_5_IRQn, 0, 0);
  HAL_NVIC_EnableIRQ(EXTI9_5_IRQn);

}

/* USER CODE BEGIN 4 */
/**
 * @brief  Period elapsed callback in non-blocking mode
 * @param  htim TIM handle
 * @retval None
 */
void HAL_TIM_PeriodElapsedCallback(TIM_HandleTypeDef *htim)
{
	/* Prevent unused argument(s) compilation warning */
	//  UNUSED(htim);

	/* NOTE : This function should not be modified, when the callback is needed,
            the HAL_TIM_PeriodElapsedCallback could be implemented in the user file
	 */
	if (htim == &htim11){
		saturn_gamepad();
	}
}

/**
 * @brief  The function is a callback about HID Data events
 *  @param  phost: Selected device
 * @retval None
 */
void USBH_HID_EventCallback(USBH_HandleTypeDef *phost)
{
	/* Prevent unused argument(s) compilation warning */
	//  UNUSED(phost);
	HID_TypeTypeDef			hid_type = USBH_HID_GetDeviceType(phost);

	// Get HID Mouse Information
	if(hid_type == HID_MOUSE){
		mouse_info = USBH_HID_GetMouseInfo(phost);
		pc88_usb_mode = PC88_MOUSE;
		memset( pc88_data, 0x00, sizeof(pc88_data) );
		set_usb_mouse();
	}
	// Get HID Keyboard Information
	else if(hid_type == HID_KEYBOARD){
		keyb_info = USBH_HID_GetKeybdInfo(phost);
		pc88_usb_mode = PC88_KEYBOARD;
		memset( pc88_data, 0x00, sizeof(pc88_data) );
		set_usb_keycode();
	}
	else{
		pc88_usb_mode = PC88_GAMEPAD;
	}
}

static void set_usb_mouse(void)
{
	uint8_t mouse_x = mouse_info->x;
	uint8_t mouse_y = mouse_info->y;
	uint8_t mouse_l = mouse_info->buttons[0];
	uint8_t mouse_r = mouse_info->buttons[1];

	uint16_t button_up    = KEYB_UP;
	uint16_t button_down  = KEYB_DOWN;
	uint16_t button_left  = KEYB_LEFT;
	uint16_t button_right = KEYB_RIGHT;

	uint16_t button_a     = KEYB_SPC;
	uint16_t button_b     = KEYB_SHIFT;

	// A
	if( mouse_l ){
		pc88_data[usb_keytable[button_a].address] |=  (uint32_t)usb_keytable[button_a].bit;
	}
	// B
	if( mouse_r ){
		pc88_data[usb_keytable[button_b].address] |=  (uint32_t)usb_keytable[button_b].bit;
	}

	// UP
	if( mouse_y < MOUSE_UP && mouse_y > MOUSE_CENTER ){
		pc88_data[usb_keytable[button_up].address] |=  (uint32_t)usb_keytable[button_up].bit;
	}
	// DOWN
	else if( mouse_y > MOUSE_DOWN && mouse_y < MOUSE_CENTER ){
		pc88_data[usb_keytable[button_down].address] |=  (uint32_t)usb_keytable[button_down].bit;
	}
	// LEFT
	if( mouse_x < MOUSE_UP && mouse_x > MOUSE_CENTER ){
		pc88_data[usb_keytable[button_left].address] |=  (uint32_t)usb_keytable[button_left].bit;
	}
	// RIGHT
	else if( mouse_x > MOUSE_DOWN  && mouse_x < MOUSE_CENTER ){
		pc88_data[usb_keytable[button_right].address] |=  (uint32_t)usb_keytable[button_right].bit;
	}
}

static void set_usb_keycode(void)
{
	// shift
	if( keyb_info->lshift || keyb_info->rshift ){
		pc88_data[usb_keytable[KEYB_SHIFT].address] |= ((uint32_t)usb_keytable[KEYB_SHIFT].bit);
	}
	// ctrl
	if( keyb_info->lctrl || keyb_info->rctrl ){
		pc88_data[usb_keytable[KEYB_CTRL].address] |= ((uint32_t)usb_keytable[KEYB_CTRL].bit);
	}
	// alt - Kana
	if( keyb_info->lalt || keyb_info->ralt ){
		pc88_data[usb_keytable[KEYB_KANA].address] |= ((uint32_t)usb_keytable[KEYB_KANA].bit);
	}
	// gui - Graph
	if( keyb_info->lgui || keyb_info->rgui ){
		pc88_data[usb_keytable[KEYB_GRAPH].address] |= ((uint32_t)usb_keytable[KEYB_GRAPH].bit);
	}

	// USBキーボードの押されているキー
	for( uint32_t loop = 0; loop < 6; loop++ ){
		uint8_t		usb_code = keyb_info->keys[loop];
		if( usb_code != 0 ){
			pc88_data[usb_keytable[usb_code].address] |= ((uint32_t)usb_keytable[usb_code].bit);
		}
	}
}
static void saturn_gamepad(void)
{
	static uint32_t prev_mode_gamepad = GAMEPAD_MODE1;
	static uint8_t  prev_led = 0;
	uint8_t	        led = 1;
	uint32_t        led_pin = led_list[mode_gamepad];

	uint32_t		ss_rxyz = 0;
	uint32_t		ss_sacb = 0;
	uint32_t		ss_rldu = 0;
	uint32_t		ss_l100 = 0;

	uint16_t button_up    = gamepad_map[mode_gamepad][BUTTON_UP];
	uint16_t button_down  = gamepad_map[mode_gamepad][BUTTON_DOWN];
	uint16_t button_left  = gamepad_map[mode_gamepad][BUTTON_LEFT];
	uint16_t button_right = gamepad_map[mode_gamepad][BUTTON_RIGHT];
	uint16_t button_a     = gamepad_map[mode_gamepad][BUTTON_A];
	uint16_t button_b     = gamepad_map[mode_gamepad][BUTTON_B];
	uint16_t button_c     = gamepad_map[mode_gamepad][BUTTON_C];
	uint16_t button_s     = gamepad_map[mode_gamepad][BUTTON_S];
	uint16_t button_lr    = gamepad_map[mode_gamepad][BUTTON_LR];

	// Swap Button
	if( GPIOA->IDR & SWITCH_Pin ){
		button_a     = gamepad_map[mode_gamepad][BUTTON_C];
		button_b     = gamepad_map[mode_gamepad][BUTTON_B];
		button_c     = gamepad_map[mode_gamepad][BUTTON_A];
	}

	HAL_GPIO_WritePin(GPIOB, SS_TH_Pin, GPIO_PIN_RESET);
	HAL_GPIO_WritePin(GPIOB, SS_TR_Pin, GPIO_PIN_SET);
	SS_DELAY;
	ss_rldu = (uint32_t)((~(GPIOB->IDR) & 0xF000) >> 12 );
	SS_WAIT;

	HAL_GPIO_WritePin(GPIOB, SS_TH_Pin, GPIO_PIN_SET);
	HAL_GPIO_WritePin(GPIOB, SS_TR_Pin, GPIO_PIN_RESET);
	SS_DELAY;
	ss_sacb = (uint32_t)((~(GPIOB->IDR) & 0xF000) >> 12 );
	SS_WAIT;

	HAL_GPIO_WritePin(GPIOB, SS_TH_Pin, GPIO_PIN_RESET);
	HAL_GPIO_WritePin(GPIOB, SS_TR_Pin, GPIO_PIN_RESET);
	SS_DELAY;
	ss_rxyz = (uint32_t)((~(GPIOB->IDR) & 0xF000) >> 12 );
	SS_WAIT;

	HAL_GPIO_WritePin(GPIOB, SS_TH_Pin, GPIO_PIN_SET);
	HAL_GPIO_WritePin(GPIOB, SS_TR_Pin, GPIO_PIN_SET);
	SS_DELAY;
	ss_l100 = (uint32_t)((~(GPIOB->IDR) & 0xF000) >> 12 );
	SS_WAIT;

	rapid_count++;
	if( rapid_count > RAPID_COUNT_MAX )		rapid_count = 0;
	memset( pc88_data, 0x00, sizeof(pc88_data) );

	if( (ss_sacb & 0x00000004) ){	// A
		pc88_data[usb_keytable[button_a].address] |=  (uint32_t)usb_keytable[button_a].bit;
		led = 0;
	}
	else if( (ss_rxyz & 0x00000004) ){	// X
		if( (rapid_count & rapid_speed) ){
			pc88_data[usb_keytable[button_a].address] |=  (uint32_t)usb_keytable[button_a].bit;
			led = 0;
		}
	}
	if( (ss_sacb & 0x00000001)){	// B
		pc88_data[usb_keytable[button_b].address] |=  (uint32_t)usb_keytable[button_b].bit;
		led = 0;
	}
	else if( (ss_rxyz & 0x00000002)){	// Y
		if( (rapid_count & rapid_speed) ){
			pc88_data[usb_keytable[button_b].address] |=  (uint32_t)usb_keytable[button_b].bit;
			led = 0;
		}
	}
	if( (ss_sacb & 0x00000002) ){	// C
		pc88_data[usb_keytable[button_c].address] |=  (uint32_t)usb_keytable[button_c].bit;
		led = 0;
	}
	else if( (ss_rxyz & 0x00000001) ){	// Z
		if( (rapid_count & rapid_speed) ){
			pc88_data[usb_keytable[button_c].address] |=  (uint32_t)usb_keytable[button_c].bit;
			led = 0;
		}
	}
	if( (ss_rldu & 0x00000008)){	// RIGHT
		pc88_data[usb_keytable[button_right].address] |=  (uint32_t)usb_keytable[button_right].bit;
	}
	if( (ss_rldu & 0x00000004)){	// LEFT
		pc88_data[usb_keytable[button_left].address] |=  (uint32_t)usb_keytable[button_left].bit;
	}
	if( (ss_rldu & 0x00000002) ){	// DOWN
		pc88_data[usb_keytable[button_down].address] |=  (uint32_t)usb_keytable[button_down].bit;
	}
	if( (ss_rldu & 0x00000001) ){	// UP
		pc88_data[usb_keytable[button_up].address] |=  (uint32_t)usb_keytable[button_up].bit;
	}
	if( (ss_sacb & 0x00000008)){	// START
		pc88_data[usb_keytable[button_s].address] |=  (uint32_t)usb_keytable[button_s].bit;
	}

	if( (ss_rxyz & 0x00000008) && (ss_l100 & 0x00000008) ){		// L&R
		pc88_data[usb_keytable[button_lr].address] |=  (uint32_t)usb_keytable[button_lr].bit;
	}
	else if( (ss_rxyz & 0x00000008) ){		// R
		rapid_speed = RAPID_FAST;
	}
	else if( (ss_l100 & 0x00000008) ){		// L
		rapid_speed = RAPID_SLOW;
	}

	// Remap USB Interface
	if ( (pc88_usb_mode == PC88_KEYBOARD) && (keyb_info != NULL) ){
		set_usb_keycode();
	}
	else if ( (pc88_usb_mode == PC88_MOUSE) && (mouse_info != NULL) ){
		set_usb_mouse();
	}

	// Change GAMEPAD MODE
	// SELECT
	if( (ss_sacb & 0x00000008) ){
		if( (ss_rldu & 0x00000004)){		// LEFT
			mode_gamepad = GAMEPAD_MODE1;
		}
		else if( (ss_rldu & 0x00000008)){	// RIGHT
			mode_gamepad = GAMEPAD_MODE2;
		}
		else if( (ss_rldu & 0x00000001) ){	// UP
			mode_gamepad = GAMEPAD_MODE3;
		}
		else if( (ss_rldu & 0x00000002) ){	// DOWN
			mode_gamepad = GAMEPAD_MODE4;
		}
		if(	prev_mode_gamepad != mode_gamepad){
			GPIOB->BSRR = (LED1_Pin | LED2_Pin | LED3_Pin | LED4_Pin) << 16;
			led_pin = led_list[mode_gamepad];
			GPIOB->BSRR = led_pin;
			prev_led = 1;
		}
	}

	if( prev_led != led ){
		if( led ){
			GPIOB->BSRR = led_pin;
		}
		else{
			GPIOB->BSRR = (LED1_Pin | LED2_Pin | LED3_Pin | LED4_Pin) << 16;
		}
	}

	prev_led = led;
	prev_mode_gamepad = mode_gamepad;
}
/* USER CODE END 4 */

/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
void Error_Handler(void)
{
  /* USER CODE BEGIN Error_Handler_Debug */
  /* User can add his own implementation to report the HAL error return state */
  __disable_irq();
  while (1)
  {
  }
  /* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{
  /* USER CODE BEGIN 6 */
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */
